# Import our functions:
from TumorDecon.data_utils import *
from TumorDecon.cibersort import *
from TumorDecon.DeconRNASeq import *
from TumorDecon.ssGSEA import *
from TumorDecon.SingScore import *

def tumor_deconvolve(mixture_data, method, patient_IDs='ALL', cell_signatures=None, up_genes=None, down_genes=None, args={}):
    """
        For each patient, given mixture data, cell signatures, and a deconvolution
            method, return frequencies/ranks for each cell type
        Inputs:
            - mixture_data: pandas df of rna gene expression data.
                Rows are genes (indexed by 'Hugo_Symbol') and columns are patients
            - method: Method for tumor deconvolution. Must be either 'CIBERSORT', 'DeconRNASeq', 'ssGSEA', or 'SingScore'
            - patient_IDs: list of patient IDs to run DeconRNASeq for.
                Alternatively, can use the string 'ALL' to run for all patients
            - cell_signatures: pandas df of Signature gene expression values for given cell types.
                Rows are genes (indexed by 'Hugo_Symbol') and columns are cell types
                *** Required for 'CIBERSORT' and 'DeconRNASeq' methods ***
                    (ignored for 'ssGSEA' and 'SingScore' methods)
            - up_genes: dictionary. Keys are cell types, vaues are a list of up-regulated genes (hugo symbols) for
                the given cell type
                *** Required for 'ssGSEA' and 'SingScore' methods ***
                    (ignored for 'CIBERSORT' and 'DeconRNASeq' methods)
            - down_genes: dictionary. Keys are cell types, vaues are a list of down-regulated genes (hugo symbols) for
                the given cell type
                *** Optional for 'SingScore' method ***
                    (ignored for 'CIBERSORT', 'DeconRNASeq', and 'ssGSEA' methods)
            - args: Dictionary of arguments to pass on to the given method
        Outputs:
            - x: dataframe of frequencies/scores for each specified patient
                and each cell type in the signature matrix, as solved for using
                the specified method
    """
    import pandas as pd
    from collections.abc import Mapping

    if not isinstance(args, Mapping):
        raise TypeError('args ({!r}) must be a dict '.format(args))

    # Check data:
    if isinstance(mixture_data, pd.DataFrame):
        # TO DO: Add check for proper indexing
        pass
    else:
        raise TypeError("mixture_data must be a pandas dataframe")

    if cell_signatures is not None:
        if isinstance(cell_signatures, pd.DataFrame):
            # TO DO: Make sure index matches that of mixture_data
            pass
        else:
            raise TypeError("cell_signatures must be a pandas dataframe")

    if up_genes is not None:
        if isinstance(up_genes, Mapping):
            # TO DO: Make sure index matches that of mixture_data
            pass
        else:
            raise TypeError("up_genes must be a dictionary")

    if down_genes is not None:
        if isinstance(down_genes, Mapping):
            # TO DO: Make sure index matches that of mixture_data
            pass
        else:
            raise TypeError("down_genes must be a dictionary")

    # don't need to check patient_IDs/up_genes/down_genes since check them within
        # each method

    # Apply the method:
    if method in ['CIBERSORT', 'Cibersort', 'cibersort']:
        if cell_signatures is not None:
            x = cibersort_main(mixture_data, cell_signatures, patient_IDs, args)
        else:
            raise ValueError("'cell_signatures' argument required for CIBERSORT method")

    elif method in ['DeconRNASeq', 'deconrnaseq', 'DeconRNAseq', 'deconRNAseq']:
        if cell_signatures is not None:
            x = DeconRNASeq_main(mixture_data, cell_signatures, patient_IDs, args)
        else:
            raise ValueError("'cell_signatures' argument required for DeconRNASeq method")

    elif method in ['ssGSEA', 'ssgsea', 'SSGSEA']:
        if up_genes is not None:
            x = ssGSEA_main(mixture_data, up_genes, patient_IDs, args)
        else:
            raise ValueError("'up_genes' argument required for ssGSEA method")

    elif method in ['SingScore', 'singscore', 'Singscore']:
        if up_genes is not None:
            x = SingScore_main(mixture_data, up_genes, down_genes, patient_IDs)
        else:
            raise ValueError("'up_genes' argument required for SingScore method")

    else:
        raise ValueError("Deconvolution Method must be one of 'CIBERSORT', 'DeconRNASeq', 'ssGSEA', or 'SingScore'")

    return x



# signature_matrix.py - functions originally authored by Shaya Akbarinejad

def create_signature_matrix(infile, cell_types, clustered=False, max_clusters=10, intermfile=None, outfile="kmeans_signature_matrix_qval.txt"):
    """
    Function:
        -takes in batch corrected datasets as input, does clustering on each cell-type
        (with the silhoette method), does differential expression analysis to get the genes that
        are significantly expressed among one cluster (using adjusted p value) and outputs signature matrix
    Inputs:
        - infile: string. Path to text file of batch corrected datasets (output of "batch_correct_datasets()" function)
        - cell_types: List of cell types to include in signature matrix
        - clustered: boolean. Whether the infile is pre-clustered or not.
        - max_clusters: int. Maximum value of K to ty for k-means clustering.
            Ignored if clustered=True
        - intermfile: string. Where to save intermediate step of clustered batch corrected datasets.
            Ignored if clustered=True
            Default is infile+"_clustered.txt"
        - outfile: string. Where to save output files to.
            Default is "kmeans_signature_matrix_qval.txt"
    Outputs:
        - Signature Matrix (as named by "outfile")
        - (if clustered=False & intermfile != None) text file of Clustered (via silhoette method) Batch Corrected Datasets
    """

    import pandas as pd

    print("Reading batch-corrected dataset file %s..." % infile)

    # Cluster each cell if not already clustered
    if not clustered:
        all_cells = pd.read_csv(infile, sep="\t")
        all_cells.set_index(list(all_cells)[0], inplace=True)
        print("Clustering batch-corrected datasets...")
        clustered_cells = cluster_each_cell(cell_types, all_cells, max_clusters)
        # Save clustered results to text file (intermediate step):
        if intermfile is None:
            intermfile = infile.replace(".txt","_clustered.txt")
        print("Saving clustered data sets to %s..." % intermfile)
        clustered_cells.to_csv(intermfile, sep="\t")

    # If already clustered, simply read in clustered data file:
    if clustered:
        clustered_cells = pd.read_csv(infile, sep="\t")
        clustered_cells.set_index(list(clustered_cells)[0], inplace=True)

    # Compute signature matrix and save to 'outfile':
    clustered_cells[clustered_cells < 0] = 0
    print("Generating signature matrix from clustered batch-corrected datasets...")
    mean_of_cluster = get_mean_of_each_cluster(cell_types, clustered_cells)
    final_sign_mat = get_differentially_expr_genes(cell_types, clustered_cells, mean_of_cluster)
    print("Saving signature matrix to %s..." % outfile)
    final_sign_mat.to_csv(outfile, sep="\t")



def find_optimal_k(cell_expr, max_clusters):
    """
    Args:
        cell_expr: (n_samples, n_genes) pd DataFrame - Transposed DataFrame of cell expression
        max_clusters: integer - Specifies maximum number of of clusters to be considered
    Returns:
        k_opt: integer - optimal number of clusters based on the Silhouette method
    """
    import numpy as np
    from sklearn.cluster import KMeans
    from sklearn.metrics import silhouette_score

    sample_num = cell_expr.shape[0]
    k_vals = np.zeros(len(range(2, min(max_clusters, sample_num))))
    range_n_clusters = list(range(2, min(max_clusters, sample_num)))
    for i, n_clusters in enumerate(range_n_clusters):
        clusterer = KMeans(n_clusters=n_clusters, random_state=10, n_init=10, n_jobs=-1, tol=1e-3)
        cluster_labels = clusterer.fit_predict(cell_expr)
        silhouette_avg = silhouette_score(cell_expr, cluster_labels)
        k_vals[i] = silhouette_avg

    k_opt = np.argmax(k_vals) + 2
    print("Optimal K = %d" % k_opt)
    return k_opt


def cluster_each_cell(cells, cell_expr, max_clusters):
    """
    Args:
        cells: list of strings - Specifies what type of cells we have
        cell_expr: (n_genes, n_samples) pd DataFrame - DataFrame of cell expression
    Returns:
        out_df: (n_genes, n_samples) pd DataFrame - clustered cell expression. each column shows the cell type and the
                cluster number that sample belongs to, e.g. CD4_subtype_1.
    """
    import pandas as pd
    from sklearn.cluster import KMeans

    out_df = pd.DataFrame(index=cell_expr.index)
    for cell in cells:
        one_cell_expr = cell_expr.loc[:, cell_expr.columns.str.contains(cell)]

        assert(one_cell_expr.shape[1] != 0), cell + " is not present in the file"

        print("%s: Finding optimal K for K-means (max # of clusters = %d)" % (cell, max_clusters))
        k_opt = find_optimal_k(one_cell_expr.T, max_clusters)
        clusterer = KMeans(n_clusters=k_opt, random_state=10, n_init=10, n_jobs=-1, tol=1e-3)
        cluster_labels = clusterer.fit_predict(one_cell_expr.T)

        one_cell_expr.columns = [cell + "_subtype_" + str(cluster_labels[i] + 1) + "." + str(i) for i in range(len(cluster_labels))]
        out_df[list(one_cell_expr)] = one_cell_expr
    out_df = out_df.rename_axis("gene")
    return out_df


def get_cluster_names(cells, cell_expr):
    """
    If the clustered dataset is imported from an external file, this function gets the number of clusters of
    each cell-type in the "cell" list, and assign name to each cluster in the format of cellName_subtype_clusterNum

    Args:
        cells: list of strings - Specifies what type of cells we have
        cell_expr: (n_genes, n_samples) pd DataFrame - Transposed DataFrame of cell expression
    Returns:
        cluster_name_list: list of strings - list of all cells and clusters in the file in the format of CD4_subtype_1.
    """
    import pandas as pd
    import numpy as np

    cluster_name_list = []
    for cell in cells:
        cell_type_list = list(cell_expr.loc[:, cell_expr.columns.str.contains(cell)])

        assert(len(cell_type_list) != 0), cell + " is not present in the file"

        cell_type_list = [cell_name[:cell_name.find(".")] for cell_name in cell_type_list]
        num_cell_cluster = len(np.unique(cell_type_list))
        for i in range(num_cell_cluster):
            cluster_name_list.append(cell + "_subtype_" + str(i+1))

    return cluster_name_list


def get_mean_of_each_cluster(cells, cell_expr):
    """
    Args:
        cells: list of strings - Specifies what type of cells we have
        cell_expr: (n_genes, n_samples) pd DataFrame - Transposed DataFrame of cell expression
    Returns:
        mean_df: (n_genes, n_clusters) pd Dataframe - mean value of each cluster across all samples of that cluster.
    """
    import pandas as pd

    cluster_name_list = get_cluster_names(cells, cell_expr)
    mean_df = pd.DataFrame(index=cell_expr.index)
    for cluster_name in cluster_name_list:
        one_cell_mean = cell_expr.loc[:, cell_expr.columns.str.contains(cluster_name)].mean(axis=1)
        mean_df[cluster_name] = one_cell_mean
    return mean_df


def detect_highly_expr_genes(one_cluster, other_clusters):
    """
    Args:
        one_cluster: (n_genes, n_samples) pd DataFrame- A single cluster with its samples
        other_clusters: (n_genes, n_samples) pd DataFrame - Other clusters!
    Returns:
        list of strings - returns the first 100 genes that are highly expressed in one_cluster, but lowly expressed in
        other_clusters
    """
    import pandas as pd
    import numpy as np

    one_cluster = np.log2(one_cluster + 1)
    other_clusters = np.log2(other_clusters + 1)
    one_cluster_min = one_cluster.min(axis=1)
    other_clusters_mean = other_clusters.mean(axis=1)
    diff = one_cluster_min - other_clusters_mean
    sub_df = diff[diff > 1]

    sub_df = sub_df.sort_values(ascending=False)
    if len(sub_df) > 100:
        sub_df = sub_df.iloc[:100]
    return list(sub_df.index)


def get_differentially_expr_genes(cells, cell_expr, mean_df, p_val_thresh=0.05):
    """
    Args:
        cells: list of strings - Specifies what type of cells we have
        cell_expr: (n_genes, n_samples) pd DataFrame - Transposed DataFrame of cell expression
        mean_df: (n_genes, n_clusters) pd Dataframe - mean value of each cluster across all samples of that cluster.
        p_val_thresh: float - Threshold for adjuster p value
    Returns:
        signature matrix - pd DataFrame
    """
    import pandas as pd
    import numpy as np
    import scipy.stats as stats
    import statsmodels.stats.multitest as multi

    cluster_name_list = get_cluster_names(cells, cell_expr)
    final_selected_genes = []
    for cluster_name in cluster_name_list:
        one_cell_expr = cell_expr.loc[:, cell_expr.columns.str.contains(cluster_name)]
        other_cell_exp = cell_expr.loc[:, ~cell_expr.columns.str.contains(cluster_name)]

        _, p_val = stats.ttest_ind(one_cell_expr, other_cell_exp, axis=1, equal_var=False)
        _, q_val, _, _ = multi.multipletests(p_val)
        one_cell_expr = one_cell_expr[q_val < p_val_thresh]
        other_cell_exp = other_cell_exp[q_val < p_val_thresh]
        selected_genes = detect_highly_expr_genes(one_cell_expr, other_cell_exp)
        final_selected_genes += selected_genes
    final_selected_genes = np.unique(final_selected_genes)

    return mean_df.loc[final_selected_genes]



# hugoify.py
# Author: Nathan Grant

import requests
from bs4 import BeautifulSoup
import sys
import pandas as pd
import numpy as np
import re
from urllib.parse import urlparse
from multiprocess import Pool



class Hugoify:
    def __init__(self, data, num_workers=7):
        ## File destination must be of type excel
        self.num_workers = 7
        if isinstance(data, str):
            self.df = pd.read_excel(data)
        elif isinstance(data, pd.DataFrame):
            self.df = data
        else:
            raise Exception("Must be a DataFrame object or a path to an excel file")
    def get_missing_ids(self):
        ## First argument: Make function that inputs a enterez id and spits out a hugo id
        ## Second argument: Make a list of all of the missing
        missing = []
        values = self.df.values
        for i in range(len(values)):  #
            if type(values[i][0]) is type(3.6) or values[i][0] == "nan":
                enterez = values[i][1]
                missing.append({'index':i, 'enterez':enterez})

        p = Pool(self.num_workers)
        hugo_ids = p.map(get_id, missing)
        p.terminate()
        p.join()

        ## Change Hugo Ids to new ids
        number_found = 0
        for i in range(len(hugo_ids)):
            self.df.iloc[missing[i]['index'],0] = hugo_ids[i]
            if hugo_ids[i] != "nan":
                number_found+=1
        print("Found ", number_found, "/", len(missing), "of the missing values")
        return self.df

def hugo(df, num_workers=7):
    h = Hugoify(df, num_workers=num_workers)
    return h.get_missing_ids()

def get_id(missing):
    from urllib.parse import urlparse
    import requests
    from bs4 import BeautifulSoup
    import numpy as np
    URL = 'https://www.ncbi.nlm.nih.gov/gene/?term='

    def get_text(page):
        soup = BeautifulSoup(page.content, 'html.parser')

        # kill all script and style elements
        spans = soup.findAll("dd", {"class": "noline"})
        if len(spans) ==0:
            return 'nan'
        hugo = str(spans[0].contents[0])
        if hugo == "" or hugo == None or len(hugo) == 0:
            return 'nan'

        return hugo

    def get_hugo(URL):
        headers = requests.utils.default_headers()
        headers.update({
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.61 Safari/537.36',
            'referer': 'https://www.google.com/search?q=genecards+729&rlz=1C1CHBF_enUS776US776&oq=genecards+729&aqs=chrome..69i57j69i64j69i60l3.9462j0j7&sourceid=chrome&ie=UTF-8'
        })
        page = requests.get(URL, headers=headers)
        if page.status_code == 200:
            return get_text(page)
        else:
            return 'nan'

    newURL = newURL = urlparse(URL+str(missing['enterez']))
    value=get_hugo(newURL.geturl())
    return value

if __name__ == "__main__":
    pass



# batch_correction.py - functions originally authored by Shaya Akbarinejad

def batch_correct_datasets(data_loc, cell_file_dict, outfile="batch_corrected_datasets.txt"):
    """
    Performs batch correction on raw data files. All files must be stored in the same folder.
    Input:
        - data_loc: string. Path to folder that holds all data files
        - cell_file_dict: Dictionary. Keys = Cell Types, Values = paths to all raw files
            containing data for the given cell type
        - outfile: string. Name to save the batch corrected datasets file under.
            Default is "batch_corrected_datasets.txt"
    Output:
        - Single file 'batch_corrected_datasets.txt' containing all batch corrected
            data for all cell types in the input dictionary. Has shape (n_genes, n_samples)
    """
    import pandas as pd

    print("Performing batch correction...")

    batch_corrected_list = []
    for cell in cell_file_dict:
        file_list = cell_file_dict[cell]
        if len(file_list) > 1:
            batch_corrected_list.append(remove_batch_effect(file_list, cell, data_loc))
        else:
            file_name = file_list[0]
            data = pd.read_csv(data_loc+file_name, sep="\t", header=0)
            data.set_index(list(data)[0], inplace=True)
            cell_type_data = data.loc[:, data.columns.str.contains(cell)]
            assert(cell_type_data.shape[1] != 0), cell + " is not present in " + file_name
            cell_type_data = cell_type_data.loc[~cell_type_data.index.duplicated(keep='first')]
            batch_corrected_list.append(cell_type_data)

    batch_corrected_df = pd.concat(batch_corrected_list, axis=1, sort=True)
    batch_corrected_df.dropna(inplace=True)

    print("Saving batch corrected datasets to %s..." % outfile)
    batch_corrected_df.to_csv(outfile, sep="\t")
    # return batch_corrected_df


def differentiate_same_col_names(input_df):
    """
    There might be multiple columns of the same name. This results in Combat function to raise an error. This function
    makes sure that each column has a unique name.

    Args:
        input_df: pd DataFrame with repetitive column names
    Returns:
        pd DataFrame with unique column names, the same size as input_df
    """
    col_names = list(input_df)
    col_list = []
    col_dict = {}
    for col_name in col_names:
        if col_name in col_dict:
            col_dict[col_name] += 1
            col_list.append(col_name + "_" + str(col_dict[col_name]))
        else:
            col_dict[col_name] = 0
            col_list.append(col_name)
    input_df.columns = col_list
    return input_df


def remove_batch_effect(files_list, cell_type, address):
    """
    Args:
        files_list: List of strings: List of names of files that include cell_type
        cell_type: string:  Name of cell type we are looking for
    Returns:
        batch_corrected_data: (n_genes, n_samples) pd Dataframe:  Combined gene expression of cell types with removed
        batch effect
    """
    import pandas as pd
    from combat.pycombat import pycombat

    batch = pd.Series()
    data_dict = []
    for (i, file_name) in enumerate(files_list):
        data = pd.read_csv(address+file_name, sep="\t", header=0)
        data.set_index(list(data)[0], inplace=True)
        cell_type_data = data.loc[:, data.columns.str.contains(cell_type)]
        assert(cell_type_data.shape[1] != 0), cell_type + " is not present in " + file_name

        temp_batch = pd.Series([i for _ in range(len(list(cell_type_data)))], index=list(cell_type_data))
        batch = batch.append(temp_batch)
        cell_type_data = cell_type_data.loc[~cell_type_data.index.duplicated(keep='first')]
        data_dict.append(cell_type_data)
    data_combined = pd.concat(data_dict, axis=1, sort=True)
    data_combined = data_combined.apply(lambda row: row.fillna(row.mean()), axis=1)
    data_combined.dropna(inplace=True)
    var = data_combined.var(axis=1)
    data_combined = data_combined[var > 0]
    data_combined = differentiate_same_col_names(data_combined)
    batch.index = list(data_combined)
    batch_corrected_data = pycombat(data_combined, batch)
    return batch_corrected_data



# ssGSEA.py

def ssGSEA(mix_data, up_genes, alpha=1.0, ties_method="average"):
    """
    Performs ssGSEA on a single sample for a single cell type and returns the enrichment score
    Inputs:
        - mix_data: Series of mixture RNA gene expression values, indexed on gene Hugo_Symbol, for a single patient
        - up_genes: a list of hugo identifiers referencing the up-regulated gene set to use to calculate enrichment score
        - alpha: Weight used in ssGSEA method
        - ties_method: see pandas.DataFrame.rank(). How to treat ties
    Output:
        - ssGSEA enrichment Score (float)
    """

    import pandas as pd
    import numpy as np

    mix_orig = mix_data.copy()#.reset_index(inplace=False)

    # Sort mixture data in descending order:
    mix_data = mix_data.rank(axis=0, method=ties_method).astype('int64') #, method='first', ascending=False)
    mix_data = mix_data.sort_values(ascending = False)

    # Sum all genes present in up_genes (signature):
    sm = 0
    # Count number of genes encountered that are not present in up_genes (signature)
    not_count = 0

    # Iterate through and compute cumulative sums:
    P_G = []
    P_NG = []
    for i, val in enumerate(mix_data):
        gene = mix_data.index[i]
        if gene in up_genes:
            sm += pow(np.abs(val), alpha)
        else:
            not_count += 1.

        P_G.append(sm)
        P_NG.append(not_count)

    # Calculate denominator for P_G using only genes in BOTH mix data and up_genes
    mix_sig_sum = np.sum(np.power(np.abs(mix_data.loc[mix_data.index.intersection(up_genes)]), alpha))

    P_G = np.array(P_G) / mix_sig_sum
    P_NG = np.array(P_NG) / not_count # denominator N_N_G_diff

    rank_diff = np.subtract(P_G, P_NG)
    rank = np.sum(rank_diff)
    return rank


def ssGSEA_main(rna_df, up_genes=None, patient_IDs='ALL', args={}):
    """
    This function does the following:
        - parses the dictionary 'args' for the arguments to pass on to the ssGSEA method.
        - If up_genes not passed in to function, uses the list of up-regulated genes provided in the ssGSEA paper
        - Runs ssGSEA() for each patient specified in 'patient_IDs' argument
        - Combines the resulting scores/ranks into a pandas dataframe (num_celltypes x num_patients)
    Inputs:
        - rna_df: pandas df of rna gene expression data.
            Rows are genes (indexed by 'Hugo_Symbol') and columns are patients
        - up_genes: dictionary. Keys are cell types, vaues are a list of up-regulated genes (hugo symbols) for
            the given cell type
        - patient_IDs: list of patient IDs to run ssGSEA on.
            Alternatively, can use the string 'ALL' to run for all patients
        - args: dictionary containing any of the following:
            - alpha: see ssGSEA()
            - ties_method: see ssGSEA()
            - print_progress: whether to print patient ID as ssGSEA iterates through
            - norm: whether or not to normalize enrichment scores by using the entire data set, as indicated
                by Barbie et al., 2009, online methods, pg. 2
    Outputs:
        - scores: pandas df. Contains ssGSEA enrichment scores for each cell type declared in 'up_genes'
            dictionary, for each patient in 'patient_IDs' list.
            Rows are indexed by cell type, columns are patient IDs
    """
    import pandas as pd
    import numpy as np
    from collections import Mapping
    from .data_utils import read_ssGSEA_up_genes

    # Read in optional arguments, or set them as default
    # Assert values are of the right data type when passed to ssGSEA() function.

    # Patient_ID must be 'ALL' or an array of specific patient IDs.
    if patient_IDs == 'ALL':
        patient_list = rna_df.columns
    elif not isinstance(patient_IDs, type([])):
        raise ValueError("patient_IDs should be either 'ALL', or an array of IDs (not a single ID)")
    else:
        patient_list = patient_IDs

    if 'alpha' in args.keys():
        alpha = args['alpha']
    else:
        alpha = 1.0
    if 'ties_method' in args.keys():
        ties_method = args['ties_method']
    else:
        ties_method = 'average'
    if 'print_progress' in args.keys():
        print_progress = args['print_progress']
    else:
        print_progress = False
    if 'norm' in args.keys():
        norm = args['norm']
    else:
        norm = False


    # Check if up_genes argument was passed in:
    if up_genes is not None:
        if not isinstance(up_genes, Mapping):
            raise ValueError("up_genes argument must be a dictionary")
    else:
        # use the up-regulated genes list that was in the ssGSEA paper:
        up_genes = read_ssGSEA_up_genes()

    # Get ranks for each cell type, for each patient:
    print("Running ssGSEA...")
    scores = pd.DataFrame()
    scores['Patient_ID'] = list(up_genes.keys())
    scores = scores.set_index(['Patient_ID'])
    for patient in patient_list:
        if patient in rna_df.columns:
            if print_progress:
                print(patient)
            patient_scores = [].copy()
            for cell_type in up_genes:
                score = ssGSEA(rna_df[patient], up_genes[cell_type], alpha=alpha, ties_method=ties_method)
                patient_scores.append(score)
            scores[patient] = patient_scores
        else:
            raise ValueError("patient_ID ({!r}) not present in rna dataframe".format(patient))

    if norm:
        range = np.max(np.array(scores)) - np.min(np.array(scores))
        scores = scores / range

    scores = scores.transpose()
    return scores



# __init__.py

from .data_utils import *
from .cibersort import *
from .DeconRNASeq import *
from .ssGSEA import *
from .SingScore import *
from .tumor_deconvolve import tumor_deconvolve
from .visualization import *
from .hugoify import *
from .batch_correction import *
from .signature_matrix import *



# singscore.py

def SingScore_main(rna_df, up_genes=None, down_genes=None, patient_IDs='ALL', args={}):
    """
    * If down_genes and up_genes both defined, will run PySingScore's implementation of bidirectional singscore.
    * If just up_genes defined, will run PySingScore's implementation of unidirectional singscore.

    Inputs:
        - rna_df: pandas df of rna gene expression data.
            Rows are genes (indexed by 'Hugo_Symbol') and columns are patients
        - up_genes: dictionary. Keys are cell types, vaues are a list of up-regulated genes (hugo symbols) for
            the given cell type
        - down_genes: dictionary. Keys are cell types, vaues are a list of down-regulated genes (hugo symbols) for
            the given cell type
        - patient_IDs: list of patient IDs to run SingScore on.
            Alternatively, can use the string 'ALL' to run for all patients
        - args is empty, but included for flexibility / eventually making the methods subclasses of each other
    Outputs:
        - scores: pandas df. Contains SingScore scores for each cell type declared in 'up_genes'
            dictionary, for each patient in 'patient_IDs' list.
            Rows are indexed by cell type, columns are patient IDs
    """
    from singscore import singscore
    import pandas as pd
    from .data_utils import read_ssGSEA_up_genes
    from collections import Mapping

    # Select a patient / list of patients to solve for their cell type frequencies:
        # Patient_ID must be 'ALL' or an array of specific patient IDs.
    if patient_IDs == 'ALL':
        patient_list = rna_df.columns
    elif not isinstance(patient_IDs, type([])):
        raise ValueError("patient_IDs should be either 'ALL', or an array of IDs (not a single ID)")
    else:
        patient_list = patient_IDs

    if up_genes is None:
        up_genes = read_ssGSEA_up_genes()
    else:
        if not isinstance(up_genes, Mapping):
            raise ValueError("up_genes argument must be a dictionary")

    if down_genes is not None:
        if not isinstance(up_genes, Mapping):
            raise ValueError("up_genes argument must be a dictionary")
        elif set(up_genes.keys()) != set(down_genes.keys()):
            raise KeyError("Keys (cell type) of up_genes and down_genes must match")

    # Select only the patients in the patient_list:
    sample_df = rna_df[patient_list]

    # Use PySingScore package to generate scores for each sample in sample_df:
    print("Running SingScore...")
    scores = pd.DataFrame()
    for cell_type in up_genes:
        if down_genes is None:
            score = singscore.score(up_gene=up_genes[cell_type], sample=sample_df)
        else:
            score = singscore.score(up_gene=up_genes[cell_type], down_gene=up_genes[cell_type], sample=sample_df)
        score.rename(columns={'total_score': cell_type}, inplace=True)
        scores = pd.concat([scores, score], axis=1)
    scores.index.name='Patient_ID'
    
    return scores



# visualization.py - originally authored by Sumeyye Su & Trang Le

import os
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
from TumorDecon.data_utils import *
from collections.abc import Mapping

# combine the cells that belong to same family:
def norm_and_combine(df):
    # Check if freqs already sum to 1, and scale them first if not:
    h=np.ones(df.shape[0])
    k=df.sum(axis=1).values
    if not np.array_equal(h,k):
        df=sum_upto_1(df)

    df = combine_celltypes(df)
    return df

# scale the frequences so they sum up to 1
def sum_upto_1(df):
    df[df<0]=0
    s=df.sum(axis=1)
    df=df.divide(s,axis=0)*1
    return df


def cell_frequency_boxplot(sample_cell_freq, title="", title_fontsize=10, save_as=None, axes_style=None, font_scale=1, xylabel_fontsize=10, rcParams={'figure.figsize':(15,7)}):
    """
     Input:
        - 'sample_cell_freq': A dataframe that include cell frequency of samples
            Rows are samples id, columns are cell names
        - 'title': string.
        - 'title_fontsize': int. Font-size for title.
        - 'save_as': string. Filename to save plot as (full or relative path). File format is
            inferred from filename's extension ('.png', '.pdf', '.eps', etc)
                If None, plot is not saved.
        - 'axes_style': a dictionary of parameters to pass to sns.set_style, to customize plot
            Valid dictionary keys can be found by running "sns.axes_style()" with no arguments
        - 'font_scale': float. Scaling for plot elements such as tick labels
        - 'xylabel_fontsize': int. Font-size for axis labels.
        - 'rcParams': a dictionary of parameters to pass as the rc argument to
            sns.set function
                Valid dictionary keys can be found at https://matplotlib.org/stable/tutorials/introductory/customizing.html
    Output:
        - cells frequency box plot in descending order
    """
    new_cell_freq = norm_and_combine(sample_cell_freq)
    b=new_cell_freq.median(axis = 0)
    b=list(zip(b.index,b))
    b = sorted(b, key=lambda x: x[-1],reverse=True)
    sorted_cells=[x[0] for x in b]
    new_cell_freq=new_cell_freq[sorted_cells]
    # Customize the plots with user-defined parameters
    sns.set(rc=rcParams, font_scale=font_scale)
    if axes_style is not None:
        if not isinstance(axes_style, Mapping):
            raise TypeError('axes_style must be a dict')
        sns.set_style("white", axes_style)
    else:
        sns.set(style="white", font_scale=font_scale)

    palette={'Macrophages':'violet','CD8 T cells':'orange','CD4 T cells':'goldenrod','Monocytes':'lightsalmon','NK cells':'olivedrab','Mast cells':'red','B cells':'darkcyan','T cells gamma delta':'dodgerblue','Dendritic cells':'gray','Plasma cells':'seagreen','Neutrophils':'navy', 'Eosinophils':'purple'}
    try:
        sns.boxplot(x="Patient_ID", y="value",order=sorted_cells, data=pd.melt(new_cell_freq),palette=palette)
    except ValueError:
        sns.boxplot(x="variable", y="value",order=sorted_cells, data=pd.melt(new_cell_freq),palette=palette)
    plt.xlabel('')
    plt.xticks(rotation=90)
    plt.ylabel('Frequency',fontsize=xylabel_fontsize)
    plt.title(title, fontsize=title_fontsize)
    plt.subplots_adjust(top=0.95,bottom=0.2)
    if save_as is not None:
        plt.savefig(save_as)
    plt.show()
    return


def cell_frequency_barchart(sample_cell_freq, title=" ", title_fontsize=10, save_as=None, axes_style=None, font_scale=1, xylabel_fontsize=10, rcParams={'figure.figsize':(15,7)}):
    """
     Input:
        - 'sample_cell_freq': A dataframe that include cell frequency of samples
            Rows are samples id, columns are cell names
        - 'title': string.
        - 'title_fontsize': int. Font-size for title.
        - 'save_as': string. Filename to save plot as (full or relative path). File format is
            inferred from filename's extension ('.png', '.pdf', '.eps', etc)
                If None, plot is not saved.
        - 'axes_style': a dictionary of parameters to pass to sns.set_style, to customize plot
            Valid dictionary keys can be found by running "sns.axes_style()" with no arguments
        - 'font_scale': float. Scaling for plot elements such as tick labels
        - 'xylabel_fontsize': int. Font-size for axis labels.
        - 'rcParams': a dictionary of parameters to pass as the rc argument to
            sns.set function
                Valid dictionary keys can be found at https://matplotlib.org/stable/tutorials/introductory/customizing.html

    Output:
        - a barchart plot that shows all cell frequency of all samples
    """
    new_cell_freq=norm_and_combine(sample_cell_freq)
    b=new_cell_freq.median(axis = 0)
    b=list(zip(b.index,b))
    b = sorted(b, key=lambda x: x[-1],reverse=True)
    sorted_cells=[x[0] for x in b]
    new_cell_freq=new_cell_freq[sorted_cells]
    # Customize the plots with user-defined parameters
    sns.set(rc=rcParams, font_scale=font_scale)
    if axes_style is not None:
        if not isinstance(axes_style, Mapping):
            raise TypeError('axes_style must be a dict')
        else:
            sns.set_style("white", axes_style)
    else:
        sns.set(style="white", font_scale=font_scale)
    palette={'Macrophages':'violet','CD8 T cells':'orange','CD4 T cells':'goldenrod','Monocytes':'lightsalmon','NK cells':'olivedrab','Mast cells':'red','B cells':'darkcyan','T cells gamma delta':'dodgerblue','Dendritic cells':'gray','Plasma cells':'seagreen','Neutrophils':'navy', 'Eosinophils':'purple'}

    color=[]
    for gene in sorted_cells:
        a=palette[gene]
        color=color+[a]

    ax=new_cell_freq.plot.barh(stacked=True,color=color,width=1)
    ax.legend(loc='right', bbox_to_anchor=(1.25,0.5),
          fancybox=True, shadow=False, ncol=1)

    plt.yticks([])
    plt.margins(0, 0)
    plt.xlabel('Frequency', fontsize=xylabel_fontsize)
    plt.ylabel('Patients', fontsize=xylabel_fontsize)
    plt.title(title, fontsize=title_fontsize)
    plt.subplots_adjust(top=0.95,left=0.05, right=0.8) # don't cut off legend
    if save_as is not None:
        plt.savefig(save_as)
    plt.show()
    return


def hierarchical_clustering(sample_cell_freq, title="", title_fontsize=10, font_scale=1, save_as=None, figsize=(20,5)):
    """
    Input:
        - 'sample_cell_freq': A dataframe that include cell frequency of samples
            Rows are samples id, columns are cell names
        - 'title': string.
        - 'title_fontsize': int. Font-size for title.
        - 'font_scale': float. Scaling for plot elements such as tick labels
        - 'save_as': string. Filename to save plot as (full or relative path). File format is
            inferred from filename's extension ('.png', '.pdf', '.eps', etc)
                If None, plot is not saved.
        - 'figsize': tuple. (x,y) size of the figure
    Output:
        - hierarchical clustered heatmap from cell frequency of samples
    """

    new_cell_freq=norm_and_combine(sample_cell_freq)

    b=new_cell_freq.median(axis = 0)
    b=list(zip(b.index,b))
    b = sorted(b, key=lambda x: x[-1],reverse=True)
    sorted_cells=[x[0] for x in b]
    new_cell_freq=new_cell_freq[sorted_cells]
    sns.set(font_scale=font_scale)
    sns.clustermap(new_cell_freq,cmap='coolwarm',row_cluster=False, figsize=figsize)
    plt.subplots_adjust(bottom=0.2)
    plt.title(title, fontsize=title_fontsize)
    if save_as is not None:
        plt.savefig(save_as)
    plt.show()
    return

def pair_plot(sample_cell_freq, title="", title_fontsize=10, font_scale=1, save_as=None, figsize=(20,20)):
    """
     Input:
        - 'sample_cell_freq': A dataframe that include cell frequency of samples
           Rows are samples id, columns are cell names
       - 'title': string.
       - 'title_fontsize': int. Font-size for title.
       - 'font_scale': float. Scaling for plot elements such as tick labels
       - 'save_as': string. Filename to save plot as (full or relative path). File format is
            inferred from filename's extension ('.png', '.pdf', '.eps', etc)
                If None, plot is not saved.
       - 'figsize': tuple. (x,y) size of the figure
     Output:
        - pairplot from cell frequency of samples
    """
    new_cell_freq=norm_and_combine(sample_cell_freq)
    b=new_cell_freq.median(axis = 0)
    b=list(zip(b.index,b))
    b = sorted(b, key=lambda x: x[-1],reverse=True)
    sorted_cells=[x[0] for x in b]
    new_cell_freq=new_cell_freq[sorted_cells]

    sns.set(font_scale=font_scale)
    p = sns.pairplot(new_cell_freq)
    p.fig.set_size_inches(*figsize)
    plt.subplots_adjust(bottom=0.1, top=1.0, left=0.05, right=0.95)
    plt.title(title, fontsize=title_fontsize)
    if save_as is not None:
        plt.savefig(save_as)
    plt.show()
    return


def stack_barchart(methods, results, true_freqs, cell_types, colors, figsize=(10,10), save_as=None):
    """
    Plot for comparing predicted frequencies of each method to true data
     Input:
        - 'methods': 1D string array.
                All methods with results to plot.
        - 'results': 3D float array (num_methods x num_cell_types x num_samples)
                All TumorDecon results from each method.
        - 'true_freqs': 2D float array (num_cell_types x num_samples)
                Experimental cell frequencies to compare results to.
        - 'cell_types': 1D string array.
                All cell_types listed in the results
        - 'colors': 1D string array of length num_cell_types.
                Enties must be valid matplotlib colors string (to associate with each given cell_type)
        - 'save_as': string. Filename to save plot as (full or relative path). File format is
            inferred from filename's extension ('.png', '.pdf', '.eps', etc)
                If None, plot is not saved.
        - 'figsize': tuple. (x,y) size of the figure
     Output:
        - pairplot from cell frequency of samples
    """
    sample_labels = np.arange(1,len(true_freqs)+1)

    fig, axs = plt.subplots(1, len(methods)+1, sharey=True, figsize=figsize)
    # Remove vertical space between axes
    fig.subplots_adjust(wspace=0.1)

    for i in range(len(methods)):
        results[methods[i]][cell_types].plot.barh(ax = axs[i], stacked=True, color=colors, width=0.9)
        axs[i].get_legend().remove()
        axs[i].set_yticklabels(sample_labels)
        axs[i].set_title(methods[i])
        axs[i].set_ylabel('Sample')
        if methods[i] in ['ssGSEA', 'singscore']:
            axs[i].set_xlabel('Score')
        else:
            axs[i].set_xlabel('Frequency')

    true_freqs[cell_types].plot.barh(ax = axs[-1], stacked=True, color=colors, width=0.9)
    axs[-1].get_legend().remove()
    axs[-1].set_yticklabels(sample_labels)
    axs[-1].set_xlabel('Frequency')
    axs[-1].set_title('Ground_truth')

    fig_leg = plt.figure(figsize=(3, 3))
    ax_leg = fig_leg.add_subplot(111)
    # add the legend from the previous axes
    ax_leg.legend(*axs[-1].get_legend_handles_labels(), loc='center')
    # hide the axes frame and the x/y labels
    ax_leg.axis('off')

    if save_as is not None:
        plt.savefig(save_as)



# data_utils.py - functions for reading in pancancer datasets and manipulating data

def get_td_Home():
    import os
    # Returns the path where this file is stored
    return os.path.realpath(__file__).strip("data_utils.py")

def read_rna_file(rna_file_path, identifier='hugo', fetch_missing_hugo=False):
    """
    Read in a cbioportal pancancer or TCGA Xena Hug txt file containing mixture gene expression data, and return a pandas dataframe.
    Columns are patients, Rows are genes
    Inputs:
        - rna_file_path: string. Relative or full path to the RNA seq file
            This file is tab seperated and includes two columns, 'Hugo_Symbol' and 'Entrez_Gene_Id', for each
            gene preceding the expression values for each patient
        - identifier: string. Determines which gene identifier to use to index the rna data.
            Must be set to either:
                - 'hugo' for Hugo Symbols
                - 'entrez' for Entrez Gene ID
        - fetch_missing_hugo: boolean. Whether to fetch missing Hugo Symbols (by Entrez Gene ID) from ncbi website
    Output:
        - pandas data frame (num_genes x num_patients) with either Hugo_Symbol or Entrez_Gene_Id as index
    """

    import pandas as pd
    from TumorDecon.hugoify import hugo

    rna = pd.read_csv(rna_file_path,  sep='\t')

    # Xena Hub:
    if "sample" in rna.columns:
        rna['Hugo_Symbol'] = rna['sample'].astype(str)
        rna = rna.set_index(['Hugo_Symbol']).drop(['sample'], axis=1)
    # CbioPortal TCGA Pan Cancer Atlas:
    elif "Hugo_Symbol" in rna.columns:
        rna[['Hugo_Symbol', 'Entrez_Gene_Id']] = rna[['Hugo_Symbol', 'Entrez_Gene_Id']].astype(str)
        if identifier == 'hugo':
            if fetch_missing_hugo:
                print("Fetching missing Hugo Symbols for genes by Entrez ID...")
                rna = hugo(rna)
            rna = rna.set_index(['Hugo_Symbol']).drop(['Entrez_Gene_Id'], axis=1)
            # Remove rows/genes that don't have a Hugo Symbol name:
            rna = rna.drop('nan',errors='ignore') # ignore error that arrises if no 'nan's
        elif identifier == 'entrez':
            rna = rna.set_index(['Entrez_Gene_Id']).drop(['Hugo_Symbol'], axis=1)
            # Remove rows/genes that don't have a Entrez ID:
            rna = rna.drop('nan')
        else:
            raise ValueError("gene identifier must be set to 'hugo' or 'entrez'")
    else:
        raise ReadError("Data Format not recognized")

    # Drop duplicate genes:
    # remove duplicates:
    # rna = rna.loc[~rna.index.duplicated(keep='first')] # argument keep = first, last, max, mean

    # Drop NA (if any):
    rna.dropna(axis=0, how='any', inplace=True)
    rna.dropna(axis=1, how='any', inplace=True)

    return rna


def download_by_name(source, type, fetch_missing_hugo=True):
    """
    Function to download TCGA data from either UCSC Xena Hub or cBioPortal given
    the desired source and cancer type, instead of a URL.
        Inputs:
            - source: either "xena" or "cbio"
            - type: specific strings. Each will be presented as options in a dropdown menu in the GUI
            - fetch_missing_hugo: boolean. Whether to fetch missing Hugo Symbols (by Entrez Gene ID) from ncbi website
        Outputs:
            - pandas df of mixture data
    """
    if source in ["xena", "Xena", "UCSC", "ucsc"]:
        type_dict = {'Acute Myeloid Leukemia':'LAML','Adrenocortical Cancer':'ACC', 'Bile Duct Cancer':'CHOL', 'Bladder Cancer':'BLCA', 'Breast Cancer':'BRCA',
                     'Cervical Cancer':'CESC', 'Colon and Rectal Cancer':'COADREAD', 'Colon Cancer':'COAD', 'Endometrioid Cancer':'UCEC', 'Esophageal Cancer':'ESCA',
                     'Glioblastoma':'GBM', 'Head and Neck Cancer':'HNSC', 'Kidney Chromophobe':'KICH', 'Kidney Clear Cell Carcinoma':'KIRC', 'Kidney Papillary Cell Carcinoma':'KIRP',
                     'Large B-cell Lymphoma':'DLBC', 'Liver Cancer':'LIHC', 'Lower Grade Glioma':'LGG', 'Lower Grade Glioma and Glioblastoma':'GBMLGG', 'Lung Adenocarcinoma':'LUAD',
                     'Lung Cancer':'LUNG', 'Lung Squamous Cell Carcinoma':'LUSC', 'Melanoma':'SKCM', 'Mesothelioma':'MESO', 'Ocular Melanomas':'UVM', 'Ovarian Cancer':'OV',
                     'Pancreatic Cancer':'PAAD', 'Pheochromocytoma and Paraganglioma':'PCPG', 'Prostate Cancer':'PRAD', 'Rectal Cancer':'READ', 'Sarcoma':'SARC', 'Stomach Cancer':'STAD',
                     'Testicular Cancer':'TGCT', 'Thymoma':'THYM', 'Thyroid Cancer':'THCA', 'Uterine Carcinosarcoma':'UCS'}
        if type in type_dict.keys():
            urlcode = type_dict[type]
        else:
            raise ValueError("type {!r} not available from UCSC Xena".format(type))
        # Download and return:
        data = download_from_xena(url="https://tcga-xena-hub.s3.us-east-1.amazonaws.com/latest/TCGA."+urlcode+".sampleMap/HiSeqV2.gz", fetch_missing_hugo=fetch_missing_hugo)
        return data
    elif source in ["cbio", "CbioPortal", "cbioportal", "cBioPortal"]:
        type_dict = {'Acute Myeloid Leukemia':'laml','Adrenocortical Carcinoma':'acc', 'Bladder Urothelial Carcinoma':'blca', 'Brain Lower Grade Glioma':'lgg', 'Breast Invasive Carcinoma':'brca',
                     'Cervical Squamous Cell Carcinoma':'cesc', 'Cholangiocarcinoma':'chol', 'Colorectal Adenocarcinoma':'coadread', 'Diffuse Large B-Cell Lymphoma':'dlbc', 'Esophageal Adenocarcinoma':'esca',
                     'Glioblastoma Multiforme':'gbm', 'Head and Neck Squamous Cell Carcinoma':'hnsc', 'Kidney Chromophobe':'kich', 'Kidney Renal Clear Cell Carcinoma':'kirc', 'Kidney Renal Papillary Cell Carcinoma':'kirp',
                     'Liver Hepatocellular Carcinoma':'lihc', 'Lung Adenocarcinoma':'luad', 'Lung Squamous Cell Carcinoma':'lusc', 'Mesothelioma':'meso', 'Ovarian Serous Cystadenocarcinoma':'ov', 'Pancreatic Adenocarcinoma':'paad',
                     'Pheochromocytoma and Paraganglioma':'pcpg', 'Prostate Adenocarcinoma':'prad', 'Sarcoma':'sarc', 'Skin Cutaneous Melanoma':'skcm', 'Stomach Adenocarcinoma':'stad', 'Testicular Germ Cell Tumors':'tgct',
                     'Thymoma':'thym', 'Thyroid Carcinoma':'thca', 'Uterine Carcinosarcoma':'ucs', 'Uterine Corpus Endometrial Carcinoma':'ucec', 'Uveal Melanoma':'uvm'}
        if type in type_dict.keys():
            urlcode = type_dict[type]
        else:
            raise ValueError("type {!r} not available from CbioPortal".format(type))
        # Download and return:
        data = download_from_cbio(url="https://cbioportal-datahub.s3.amazonaws.com/"+urlcode+"_tcga_pan_can_atlas_2018.tar.gz", fetch_missing_hugo=fetch_missing_hugo)
        return data
    else:
        raise ValueError("source ({!r}) must be 'xena' or 'cbioportal'".format(source))



def download_from_cbio(url="https://cbioportal-datahub.s3.amazonaws.com/uvm_tcga_pan_can_atlas_2018.tar.gz", save_location=get_td_Home()+"data/downloaded/", delete_after=False, fetch_missing_hugo=True):
    """
    Function to download data directly from cbioportal and read it in as a pandas df.
        Inputs:
            - url: url to data
            - save_loc: where to save the data. Default is in TumorDecon directory, under data/downloaded
            - delete_after: not implemented yet. Whether to delete data after reading
            - fetch_missing_hugo: boolean. Whether to fetch missing Hugo Symbols (by Entrez Gene ID) from ncbi website
        Outputs:
            - pandas df of mixture data
    """
    import wget
    import tarfile
    import os

    # Check if file already downloaded, if not, download it:
    file = save_location+url.replace('https://cbioportal-datahub.s3.amazonaws.com/','')
    if not os.path.exists(file):
        # Download file and save it locally:
        print("Downloading data from cbioportal...")
        wget.download(url, save_location)
    # Unzip if applicable
    folder = file.replace(".tar.gz","")
    if file.endswith("tar.gz") and not os.path.exists(folder):
        tar = tarfile.open(file, "r:gz")
        tar.extract(folder.split("/")[-1]+"/data_RNA_Seq_v2_expression_median.txt", path=save_location)
        # tar.extractall(folder.strip(".tar.gz"))
        tar.close()
    # Read in data:
    return read_rna_file(folder+"/data_RNA_Seq_v2_expression_median.txt", fetch_missing_hugo=fetch_missing_hugo)


def download_from_xena(url="https://tcga-xena-hub.s3.us-east-1.amazonaws.com/latest/TCGA.UCS.sampleMap/HiSeqV2.gz", save_location=get_td_Home()+"data/downloaded/", delete_after=False, fetch_missing_hugo=True):
    """
    Function to download data directly from TCGA Xena Hub and read it in as a pandas df.
        Inputs:
            - url: url to data
            - save_loc: where to save the data. Default is in TumorDecon directory, under data/downloaded
            - delete_after: not implemented yet. Whether to delete data after reading
            - fetch_missing_hugo: boolean. Whether to fetch missing Hugo Symbols (by Entrez Gene ID) from ncbi website
        Outputs:
            - pandas df of mixture data
    """
    import wget
    import gzip
    import os

    # Since all cancer types have the same filename, re-download every time:
    zipfile = save_location+url.split('/')[-1]
    if os.path.exists(zipfile):
        os.remove(zipfile)
    # Download file and save it locally:
    print("Downloading data from Xena Hub...")
    wget.download(url, save_location)
    # pandas read_csv can handle the zipped file - no need to unzip
    # Read in data:
    logdf = read_rna_file(zipfile, fetch_missing_hugo=fetch_missing_hugo)
    # Transform from log2(x+1):
    df = 2.**logdf - 1
    return df


# # OBSOLETE! Use read_sig_file() with no arguments to read in LM22.txt instead
# def read_lm22_file(file_path=get_td_Home()+"data/LM22.txt"):
#     """
#     Read in the LM22 (or LM6) signature matrix file (containing signature gene expression data for 22 different cell types)
#     and return a pandas dataframe
#     Inputs:
#         - file_path: string. Relative or full path to the LM22/LM6 signature matrix file
#             This file is tab seperated. Columns are cell types, Rows are gene Hugo Symbols.
#     Output:
#         - pandas data frame. Columns are cell types, Rows are genes, indexed by Hugo Symbol.
#     """
#
#     import pandas as pd
#
#     print("WARNING: 'read_lm22_file()'' depreciated. Use 'read_sig_file()' with no arguments to read in 'LM22.txt' signatures")
#
#     lm22 = pd.read_csv(file_path, sep='\t')
#     # lm22['Hugo_Symbol'] = lm22['Gene symbol']
#     # lm22 = lm22.drop(['Gene symbol'], axis = 1)
#     new_column_names = lm22.columns.tolist()
#     new_column_names[0] = 'Hugo_Symbol'
#     lm22.columns = new_column_names
#     lm22 = lm22.set_index(['Hugo_Symbol'])
#
#     return lm22


def read_sig_file(file_path=get_td_Home()+"data/LM22.txt", geneID="Hugo_Symbol"):
    """
    Reads in a signature matrix file (containing signature gene expression data for
    a number of different cell types), converts the gene identifiers to Hugo Symbols,
    and returns a pandas dataframe.
        ** IF NO FILE PATH GIVEN, LM22 IS ASSUMED **
    Inputs:
        - file_path: string. Relative or full path to signature matrix file.
            This file is tab seperated. Columns are cell types, Rows are genes
            If no file_path given, LM22 is assumed.
        - geneID: string in ["Hugo_Symbol", "Ensembl_Gene_ID", "Entrez_Gene_ID"].
            Describes how genes are labeled in the signature matrix file
    Output:
        - pandas data frame. Columns are cell types, Rows are genes, indexed by Hugo Symbol.
    """
    import pandas as pd

    # Read file:
    sig = pd.read_csv(file_path, sep='\t')
    # Get name of column with gene identifiers:
    gene_col = list(sig)[0]

    to_be_dropped = []
    if geneID == "Hugo_Symbol":
        sig['Hugo_Symbol'] = sig[gene_col]

    elif geneID == 'Ensembl_Gene_ID':
        sig['Hugo_Symbol'], to_be_dropped = convert_emsembl(sig[gene_col])

    elif geneID == 'Entrez_Gene_ID':
        raise ValueError("Conversion from Entrez Gene ID to Hugo Symbol not yet supported in TumorDecon")

    else:
        raise ValueError("geneID ({!r}) must be set to 'Hugo_Symbol', 'Ensembl_Gene_ID', or 'Entrez_Gene_ID'".format(geneID))

    # Drop old gene ID column, rename columns, and set Index to Hugo Symbol:
    sig = sig.drop([gene_col], axis = 1)
    new_column_names = sig.columns.tolist()
    sig.columns = new_column_names
    sig = sig.set_index(['Hugo_Symbol'])

    # Drop columns without a Hugo Symbol:
    sig = sig.drop(to_be_dropped)
    return sig

def convert_emsembl(ensembl):
    """
    Input:
        - ordered array of Ensembl Gene IDs to be converted to Hugo Symbols
    Outputs:
        - hugo: array. Ordered array of Hugo Symbols correponsing to the Ensembl Gene IDs
        - to_be_dropped: array. List of genes without a correponsing hugo symbol (to be dropped)
    """
    import pandas as pd
    import numpy as np

    hugo = np.array(ensembl)
    # Will drop any genes that don't have a corresponding Hugo Symbol
    to_be_dropped = []

    # Create dictionary for conversion:
    ensembl_hugo = pd.read_csv(get_td_Home()+"data/gene_ID_conversion/hugo_enseml_synonym.txt", header=0, sep="\t")
    ensembl_hugo2 = pd.read_csv(get_td_Home()+"data/gene_ID_conversion/GSE87692_Primary_HsESF_TPMs_073015.txt", header=0, sep="\t")
    ensembl_hugo2 = ensembl_hugo2[['ensemble_geneid', 'Associate Gene Name']]
    ensembl_hugo2.set_index('Associate Gene Name', inplace=True)
    ensembl_hugo.dropna(subset=['Ensembl ID(supplied by Ensembl)', 'Ensembl gene ID'], how='all', inplace=True)
    ensembl_hugo = ensembl_hugo.set_index('Approved symbol')

    for i, gene in enumerate(ensembl):
        # first search in the ensembl_hugo2 then if not found in ensembl_hugo
        is_found = ensembl_hugo2.loc[ensembl_hugo2['ensemble_geneid'] == gene]
        if len(is_found) > 0 and is_found.index[0] != "#N/A":
            hugo[i] = is_found.index[0]

        else:
            t = ensembl_hugo.loc[ensembl_hugo['Ensembl ID(supplied by Ensembl)'] == gene]
            # first search in 'Ensembl ID(supplied by Ensembl)' column, if not found, search the other col
            if len(t) > 0:
                if type(t.index[0]) == str:
                    hugo[i] = t.index[0]

            else:
                t2 = ensembl_hugo.loc[ensembl_hugo['Ensembl gene ID'] == gene]
                if len(t2) > 0:
                    if type(t2.index[0]) == str:
                        hugo[i] = t2.index[0]
                    else:
                        to_be_dropped.append(gene)
                else:
                    to_be_dropped.append(gene)
    if len(to_be_dropped) > 0:
        print("Missing Hugo Symbols for the following genes (will be dropped from signature matrix):")
        print(to_be_dropped)
    return hugo, to_be_dropped


def df_normalization(df, scaling, axis=0):
    """
    Inputs:
        - df: pandas data frame
        - scaling: string, must be either 'zscore' or 'minmax'.
        - axis: 0 or 1. Whether to apply normalization across individual columns (0) or individual rows (1)
    Output:
        - df_norm: normalized (minmax) data
    """
    import pandas as pd
    import numpy as np
    from sklearn import preprocessing
    if scaling == "zscore":
        df_norm = preprocessing.scale(df, axis=axis)
        df_norm = pd.DataFrame(df_norm)
    elif scaling == "r-zscore":
        # uses delta degrees of freedom = 1 (as R std does), instead of 0 (as sklearn does)
        mean = np.mean(df, axis=axis)
        std = np.std(df, axis=axis, ddof=1)
        df_norm = (df - mean) / std
        df_norm = pd.DataFrame(df_norm)
    elif scaling == "minmax":
        min_max_scaler = preprocessing.MinMaxScaler()
        if axis == 1:
            df_norm = df.transpose()
        else:
            df_norm = df.copy()
        df_norm = min_max_scaler.fit_transform(df_norm)
        df_norm = pd.DataFrame(df_norm)
        if axis == 1:
            df_norm = df_norm.transpose()
    # add column names back to the dataframe
    df_norm.columns = df.columns
    # add row index back to the dataframe
    df_norm.index = df.index
    return df_norm


def keep_common_genes(rna_df, sig_df):
    """
    Given two dataframes, eliminate all genes that are not present in both dataframes.
    Both dataframes must be indexed with the same gene identifier
    Inputs:
        - rna_df: pandas df. Rows are genes (indexed by 'Hugo_Symbol') and columns are patients
        - sig_df: pandas df. Rows are genes (indexed by 'Hugo_Symbol') and columns are cell types
    Output:
        - rna_df_red: pandas data frame. rna_df reduced to only contain genes also present in sig_df
        - sig_df_red: pandas data frame. sig_df reduced to only contain genes also present in rna_df
    """
    import pandas as pd

    # Eliminate genes that are not present in both rna df and signature matrix
    shared_index = pd.merge(rna_df, sig_df, how='inner', left_index=True, right_index=True)
    rna_df_red = rna_df.loc[shared_index.index]
    sig_df_red = sig_df.loc[shared_index.index]

    # Make sure genes listed in same order for both dataframes lines up:
    rna_df_red = rna_df_red.sort_index()
    sig_df_red = sig_df_red.sort_index()

    return rna_df_red, sig_df_red


def find_up_down_genes_from_sig(sig_df, down_cutoff=0.4, up_cutoff=4.0, show_plots=False):
    """
    Function divides each gene expression value in Signature Matrix (sig_df) by the median value of the
        given gene across all cell types.
    Then all genes with ratios below 'down_cutoff' are considered "down regulated genes" while all genes with
    ratios above 'up_cutoff' are considered "up regulated genes".
    Inputs:
        - sig_df: Pandas df. Signature matrix (rows are genes, columns are cell types) indexed by Hugo Symbol
        - down_cutoff: value to use for the cutoff point as to what should be considered a "down regulated gene"
        - up_cutoff: value to use for the cutoff point as to what should be considered a "up regulated gene"
        - show_plots: boolean. Whether to show a plot of the sorted signature matrix ratios for each cell type.
            Can be used to help choose the cutoff points
    Outputs:
        - up: dictionary. Keys are cell types, values are a list of up-regulated genes for that cell type
        - down: dictionary. Keys are cell types, values are a list of down-regulated genes for that cell type
    """
    import pandas as pd
    import matplotlib.pyplot as plt

    ref = sig_df.copy()
    ref['reference'] = ref.median(axis=1)
    sig_ratios = sig_df.divide(ref['reference'], axis=0)

    if show_plots:
        t = range(sig_ratios.shape[0])
        for cell in sig_ratios.columns:
            plt.figure()
            vals = sig_ratios[cell]
            vals = vals.sort_values()
            axes = plt.gca()
            axes.set_ylim([0,10])
            axes.plot(t, vals)
            plt.title(cell)
            plt.show()

    # Generate list of up-regulated and down-regulated gene sets
    up = {}
    down = {}
    for cell in sig_ratios.columns:
        down[cell] = sig_ratios[sig_ratios[cell] < down_cutoff].index.tolist()
        up[cell] = sig_ratios[sig_ratios[cell] > up_cutoff].index.tolist()

    return up, down


def get_top_ranked_genes_from_sig(sig_df, sig_size=50):
    """
    Function returns the top 'sig_size' most expressed genes in the signature matrix, for each cell type
    Inputs:
        - sig_df: Pandas df. Signature matrix (rows are genes, columns are cell types) indexed by Hugo Symbol
        - sig_size: int. Number of genes to select
    Outputs:
        - dictionary. Keys are cell types, values are a list of the "sig_size" top most expressed genes in the
        signature matrix for that cell type
    """
    import pandas as pd

    top_genes = pd.DataFrame()
    for cell in sig_df.columns:
        top_genes[cell] = sig_df[cell].sort_values(ascending=False).index[:sig_size]

    return top_genes.to_dict('list')



def variance_threshold_selector(data, threshold=0.5):
    """
    Function to fit sklearn's variance threshold feature selection method on a df where the features (genes) are
    the rows in a pandas df (instead of columns)
    Inputs:
        - data: pandas df with genes as rows and either patients/celltype as column
        - threshold: The cut-off value to use for variance threshold selection
    Outputs:
        - reduced data frame (eliminates genes with variance < threshold) across patients/celltypes
    """
    from sklearn.feature_selection import VarianceThreshold

     # Transpose to work with sklearn's implementation, which assumes features are columns:
    data = data.T
    # Fit selector:
    selector = VarianceThreshold(threshold)
    selector.fit(data)
    # Apply transformation to df:
    data_red = data[data.columns[selector.get_support(indices=True)]]
    # Transpose back so genes are rows again:
    return data_red.T


def read_ssGSEA_up_genes(file_path=get_td_Home()+'data/Gene_sets.csv'):
    """
    Function to read in csv file containing the gene sets from the ssGSEA paper
    Inputs:
        none (hard-coded path to correct csv file)
    Outputs:
        - dictionary. Keys are the 26 cell types used in ssGSEA paper, values are a list of up regulated genes
            for that cell type
    """
    import pandas as pd

    gene_sets = pd.read_csv(file_path)
    up_genes = {}
    for cell in gene_sets['Cell type'].unique():
        up_genes[cell] = gene_sets[gene_sets['Cell type'] == cell]['Symbol'].tolist()

    return up_genes

def read_geneset(file_path=get_td_Home()+"data/LM22_up_genes.csv"):
    """
        Function to read in a custom csv file containing the up or down regulated
            genes for each cell type
        Inputs:
            - file_path: string. Relative or full path to signature matrix file.
            File should have columns named for each cell type, and rows should
            contain Hugo Symbols for genes to be considered as up (or down)
            regulated for that cell type. If not all cell types have the same number of up(down)
            regulated genes, excess rows in each column should be coded as "NaN"
                If no file_path given, the up-regulated gene set derived in
                Le et al. (2020) is assumed.
        Outputs:
            - dictionary. Keys are the column names of the given csv file (cell types),
                values are a list of up (down) regulated genes for that cell type
    """
    import pandas as pd
    import numpy as np
    gene_sets = pd.read_csv(file_path)
    chosen_genes = {} # up or down genes
    for cell in gene_sets.columns.unique():
        chosen_genes[cell] = list(gene_sets[cell].dropna())
    return chosen_genes


def combine_celltypes(df, cols_to_combine=None):
    """
    Function to sum related cell types into a single column
    Inputs:
        - df: pandas dataframe. Output of td.tumor_deconvolve()
        - cols_to_combine: dictionary. Keys are the desired names of any new cell type columns,
            values are an arary of current column names to combine under the key name
                - Default = dictionary for combining common cell types from LM22
    Outputs:
        - Pandas dataframe with columns combined as specified by the dictionary.
            All unmentioned column names are left as they are
    """
    import pandas as pd

    if cols_to_combine is not None:
        from collections.abc import Mapping
        if isinstance(cols_to_combine, Mapping):
            pass
        else:
            raise TypeError("cols_to_combine must be a dictionary")
    else:
        # Use LM22 as default
        print("WARNING: No dictionary defined for combining columns... Attempting to use default dict for LM22 signatures")
        cols_to_combine = {'B cells':['B cells naive', 'B cells memory'],
                           'CD4 T cells':['T cells CD4 naive', 'T cells CD4 memory resting', 'T cells CD4 memory activated','T cells follicular helper', 'T cells regulatory (Tregs)'],
                           'CD8 T cells':['T cells CD8'],
                           'NK cells':['NK cells resting', 'NK cells activated'],
                           'Macrophages':['Macrophages M0', 'Macrophages M1', 'Macrophages M2'],
                           'Mast cells':['Mast cells resting','Mast cells activated'],
                           'Dendritic cells':['Dendritic cells resting', 'Dendritic cells activated']
                          }
    df2 = df.copy()
    for cell_type in cols_to_combine.keys():
        try:
            df2[cell_type] = df2[cols_to_combine[cell_type]].sum(axis=1)
            df2.drop(cols_to_combine[cell_type],axis=1,inplace=True)
        except KeyError as e:
            print("WARNING: Failed to combine some columns: ")
            print("KeyError: "+str(e))
            pass

    return df2



def corr_table(methods, results, cell_types, true_freqs):
    import pandas as pd
    import numpy as np
    from scipy.stats.stats import pearsonr
    from scipy.stats import spearmanr

    p_corr_per_cell = pd.DataFrame(index=cell_types, columns=methods)
    p_corr_per_sample = pd.DataFrame(index=true_freqs.index, columns=methods)
    s_corr_per_cell = pd.DataFrame(index=cell_types, columns=methods)
    s_corr_per_sample = pd.DataFrame(index=true_freqs.index, columns=methods)
    for method in methods:
        for cell in cell_types:
            p_corr_per_cell.loc[cell, method] = pearsonr(results[method][cell], true_freqs[cell])[0]
            s_corr_per_cell.loc[cell, method] = spearmanr(results[method][cell], true_freqs[cell])[0]
        for i in range(len(true_freqs)):
            p_corr_per_sample.loc[true_freqs.index[i], method] = pearsonr(results[method][cell], true_freqs[cell])[0]
            s_corr_per_sample.loc[true_freqs.index[i], method] = spearmanr(results[method][cell], true_freqs[cell])[0]
    return np.abs(p_corr_per_cell), np.abs(p_corr_per_sample), np.abs(s_corr_per_cell), np.abs(s_corr_per_sample)


def corr_mean_std(corr_per_cell, corr_per_sample):
    import pandas as pd
    import numpy as np

    corr = pd.DataFrame(data=np.zeros((corr_per_cell.shape[1], 4)), index=corr_per_cell.columns, columns=['Mean_corr_per_sample', 'Std_corr_per_sample', 'Mean_corr_per_cell', 'Std_corr_per_cell'])
    corr['Mean_corr_per_sample'] = np.mean(corr_per_sample, axis=0)
    corr['Std_corr_per_sample'] = np.std(corr_per_sample, axis=0)
    corr['Mean_corr_per_cell'] = np.mean(corr_per_cell, axis=0)
    corr['Std_corr_per_cell'] = np.std(corr_per_cell, axis=0)
    return corr



def flatten_corr_per_cell(corr_per_cell):
    import pandas as pd
    import numpy as np

    corr_per_cell2 = pd.DataFrame(data=np.zeros((corr_per_cell.shape[0]*corr_per_cell.shape[1], 3)), columns=['Method', 'Cell_type', 'Correlation'])
    methods = []
    for method in corr_per_cell.columns:
        methods.extend([method]*corr_per_cell.shape[0])
    corr_per_cell2['Method'] = methods
    corr_per_cell2['Cell_type'] = list(corr_per_cell.index)*corr_per_cell.shape[1]
    corr = []
    for method in corr_per_cell.columns:
        corr.extend(list(corr_per_cell[method]))
    corr_per_cell2['Correlation'] = corr

    return corr_per_cell2


def predicted_truth_bycell(method, method_freqs, exp_freqs, cell_types):
    import pandas as pd

    df = pd.concat([method_freqs[cell_types], exp_freqs[cell_types]])
    df['Method'] = [method]*exp_freqs.shape[0] + ['Ground truth']*exp_freqs.shape[0]
    df = pd.DataFrame(data=np.zeros((exp_freqs.shape[0]*len(cell_types), 3)),
                              columns=[method, 'Ground truth', 'Cell type'])
    method_fractions, true_fractions, cell_list = [], [], []
    for cell in cell_types:
        method_fractions.extend(list(method_freqs[cell]))
        true_fractions.extend(list(exp_freqs[cell]))
        cell_list.extend([cell]*exp_freqs.shape[0])
    df[method] = method_fractions
    df['Ground truth'] = true_fractions
    df['Cell type'] = cell_list

    return df



import setuptools

with open("README.md", "r") as fh:
    readme = fh.read()

setuptools.setup(
    name="TumorDecon",
    version="0.2.2",
    author="ShahriyariLab",
    author_email="lshahriyari@umass.edu",
    description="Deconvolution Methods for Digital Cytometry",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/ShahriyariLab/TumorDecon",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    license="MIT License",
    include_package_data=True,
    dependency_links=["git+https://github.com/kristyhoran/singscore"],
    install_requires=[
        "scikit-learn>0.20.3",
        "numpy>1.18.2",
        "pandas>1.0.3",
        "matplotlib>2.2.2",
        "scipy>1.2.1",
        "requests>2.22.0",
        "beautifulsoup4>4.6.0",
        "multiprocess>=0.70.9",
        "seaborn>0.10.0",
        "wget>=3.2",
        "combat>=0.1.7",
        "statsmodels>=0.9.0"
    ]
)



# cibersort.py

def cibersort(rna_sample, sig_df, nu=0.5, C=1.0, kernel='linear', shrinking=True):
    """
    Uses NuSVR from sklearn to solve for the cell type frequencies
    Inputs:
        - rna_sample: pandas series. A single sample of rna expression values
        - sig_df: pandas df of Signature gene expression values for given cell types.
            Rows are genes (indexed by 'Hugo_Symbol') and columns are cell types
        - nu: see sklearn's NuSVR
        - C: see sklearn's NuSVR
        - kernel: see sklearn's NuSVR
        - shrinking: see sklearn's NuSVR
    Outputs:
        - weights: NuSVR solution vector for the given sample.
            (Negative values in the solution vector are set to 0 for interpretation as cell type frequencies)
    """
    from sklearn.svm import NuSVR
    import numpy as np
    from sklearn.model_selection import GridSearchCV

    # If a numerical of nu not explicitly specified, use gridsearch to find the best nu:
    if nu == 'best':
        gridsearch = GridSearchCV(NuSVR(C=C, kernel=kernel, max_iter=-1, shrinking=shrinking), cv=5, param_grid={"nu": [0.25, 0.5, 0.75]}, scoring='neg_mean_squared_error', refit=True)
        gridsearch.fit(sig_df, rna_sample)
        nu = gridsearch.best_params_['nu']

    # Fit nuSVR with best (or specified) value of nu:
    clf = NuSVR(nu=nu, C=C, kernel=kernel, max_iter=-1, shrinking=shrinking, tol=1e-3)
    clf.fit(sig_df, rna_sample)

    # Replace negative "frequencies" with 0:
    weights = np.array(clf.coef_)[0] # equivalent to np.matmul(np.array(clf.dual_coef_), np.array(clf.support_vectors_))[0]
    weights[weights<0] = 0
    # Sum to 1 contraint:
    weights = weights / np.sum(weights)

    return weights


def cibersort_main(rna_df, sig_df, patient_IDs='ALL', args={}):
    """
    This function does the following:
        - parses the dictionary 'args' for the arguments to pass on to the CIBERSORT method.
        - eliminates genes from rna_df and sig_df that are not present in both data sets
        - Runs cibersort() for each patient specified in patient_IDs' argument
        - Combines the resulting frequencies into a pandas dataframe (num_celltypes x num_patients)
    Inputs:
        - rna_df: pandas df of rna gene expression data.
            Rows are genes (indexed by 'Hugo_Symbol') and columns are patients
        - sig_df: pandas df of Signature gene expression values for given cell types.
            Rows are genes (indexed by 'Hugo_Symbol') and columns are cell types
        - patient_IDs: list of patient IDs to run cibersort for.
            Alternatively, can use the string 'ALL' to run for all patients
        - args: dictionary containing any of the following:
            - print_progress: whether to print patient ID as cibersort iterates through
            - scaling: string, must be either 'None', 'zscore', or 'minmax'. Determines how to scale the mixture data and signature matrix before applying CIBERSORT
            - scaling_axis: 0 or 1. Whether to scale mixture data and signature matrix by normalizing each column (patient/celltype) individually (scaling_axis=0) or each row (gene) individually (scaling_axis=1).
            - nu: see sklearn's NuSVR. If nu='best', will use the nu in {0.25, 0.5, 0.75} that minimizes root mean square b/w data and prediction. If nu a float, will compute cibersort for that specific nu
            - C: see sklearn's NuSVR
            - kernel:: see sklearn's NuSVR
            - shrinking: see sklearn's NuSVR
    Outputs:
        - cell_freqs: pandas df. Contains cell type frequencies for each patient in 'patient_IDs' list.
            Rows are indexed by cell type, columns are patient IDs
    """

    import pandas as pd
    import numpy as np
    from .data_utils import keep_common_genes
    from .data_utils import df_normalization

    # Patient_ID must be 'ALL' or an array of specific patient IDs.
    if patient_IDs == 'ALL':
        patient_list = rna_df.columns
    elif not isinstance(patient_IDs, type([])):
        raise ValueError("patient_IDs should be either 'ALL', or an array of IDs (not a single ID)")
    else:
        patient_list = patient_IDs

    if 'print_progress' in args.keys():
        print_progress = args['print_progress']
    else:
        print_progress = False

    if 'scaling' in args.keys():
        scaling = args['scaling']
        if scaling not in ['None', 'none', 'zscore', 'minmax', 'r-zscore']:
            raise ValueError("scaling ({!r}) must be set to 'none', 'zscore' or 'minmax'".format(scaling))
        else:
            scaling = args['scaling']
    else: # default
        scaling = "minmax"

    if 'scaling_axis' in args.keys():
        scaling_axis = args['scaling_axis']
        if scaling_axis not in [0, 1]:
            raise ValueError("scaling_axis ({!r}) must be 0 or 1".format(scaling_axis))
    else:
        scaling_axis = 0

    if 'nu' in args.keys():
        nu = args['nu']
        if isinstance(nu, str):
            if nu != 'best':
                raise ValueError("nu ({!r}) must be either a float, or 'best'".format(nu))
    else:
        nu = 'best'

    if 'C' in args.keys():
        C = args['C']
    else:
        C = 1.0
    if 'kernel' in args.keys():
        kernel = args['kernel']
    else:
        kernel = 'linear'
    if 'shrinking' in args.keys():
        shrinking = args['shrinking']
    else:
        shrinking = True

    # eliminate genes not present in both rna and sig dfs, and ensure they list genes in the same order:
    rna_df, sig_df = keep_common_genes(rna_df, sig_df)

    # Scale data:
    if scaling in ['zscore', 'minmax', 'r-zscore']:
        sig_df = df_normalization(sig_df, scaling=scaling, axis=scaling_axis)
        rna_df = df_normalization(rna_df, scaling=scaling, axis=scaling_axis)

    # For each patient, run cibersort to get cell type frequencies, and save results to pandas df:
    cell_freqs_df = pd.DataFrame(columns = patient_list)
    cell_freqs_df['Patient_ID'] = sig_df.columns
    cell_freqs_df = cell_freqs_df.set_index(['Patient_ID'])

    print("Running CiberSort...")
    for patient in patient_list:
        if patient in rna_df.columns:
            if print_progress:
                print(patient)
            cell_freqs_df[patient] = cibersort(rna_df[patient], sig_df, nu=nu, C=C, kernel=kernel, shrinking=shrinking)
        else:
            raise ValueError("patient_ID ({!r}) not present in rna dataframe".format(patient))
            return
    cell_freqs_df = cell_freqs_df.transpose()
    return cell_freqs_df



# DeconRNASeq.py

def compute_least_squares(x, A, b):
    '''
    Inputs:
        x = vector of length n (frequency vector of length n_celltypes)
    Parameters:
        A = matrix of size p x n (signature matrix of size n_genes x n_celltypes)
        b = vector of size p (mixture data vector of length n_genes)
    Output:
        value of ||Ax-b||^2
    '''
    import numpy as np
    return np.linalg.norm(np.matmul(A,x)-b)


def compute_ridge(x, A, b, reg=1.0):
    '''
    Inputs:
        x = vector of length n (frequency vector of length n_celltypes)
    Parameters:
        A = matrix of size p x n (signature matrix of size n_genes x n_celltypes)
        b = vector of size p (mixture data vector of length n_genes)
        reg = regulation constant (lamba)
    Output:
        value of ||Ax-b||^2 + ||x||^2
    '''
    import numpy as np
    return np.linalg.norm(np.matmul(A,x)-b) + reg*np.linalg.norm(x)


def compute_lasso(x, A, b, reg=1.0):
    '''
    Inputs:
        x = vector of length n (frequency vector of length n_celltypes)
    Parameters:
        A = matrix of size p x n (signature matrix of size n_genes x n_celltypes)
        b = vector of size p (mixture data vector of length n_genes)
        reg = regulation constant (lamba)
    Output:
        value of ||Ax-b||^2 + Sum_i(|x_i|)
    '''
    import numpy as np
    return np.linalg.norm(np.matmul(A,x)-b) + reg*np.sum(np.abs(x))



def DeconRNASeq(S, m, formulation="qp", reg_constant=1.0, print_results=False, label=''):
    """
    Solves for the vector x that minimizes ||Sx-m||^2 under the additional constraints:
        Sum(x_i) = 1
        x_i >=0 for all i
    Depends on the function 'compute_least_squares(x, A, b)'
    Inputs:
        - S: numpy array (n_genes x n_celltypes), Signature matrix
        - m: numpy array (n_genes x 1), Mixture data vector for a given patient
        - formulation: string, must be either 'qp', 'ridge', or 'lasso'. Determines how to formulate the optimization problem:
            'qp': solve a QP (quadratic programming) problem min_x||Ax-b||^2 with strict condition x.sum()=1, xi>=0
            'ridge': solve a ridge regression problem min_x{||Ax-b||^2 + ||x||^2} with condition xi>=0
            'lasso': solve a lasso regression problem min_x{||Ax-b||^2 + Sum_i{|x_i|} with condition xi>=0
        - reg_constant: float, regularization constant for lasso/ridge regression
        - print_result: boolean, whether or not to print the solver (scipy.optimize) results
        - label: patient ID label. Only used if print_results = True
    Outputs:
        - xopt: optimal frequency vector found by scipy.optimize
        - (if print_results) prints the scipy.optimize output to standard output
    """
    import scipy.optimize as opt
    import numpy as np
    import warnings

    # Initialize frequencies with random numbers between 0 and 1:
    x0 = np.random.rand(S.shape[1])

    if formulation == 'qp':
        # Set bounds on x_i values to be between 0 and 1
        bnds = ((0., 1.),) * S.shape[1]

        # Define equality constraint that Sum(x_i) = 1
        cons = ({'type': 'eq',
                 'fun' : lambda x: np.array([np.sum(x)-1])})

        # Find the x that minimizes this least squares problem:
        soln = opt.minimize(compute_least_squares, x0, bounds=bnds, constraints=cons, args=(S,m))

    else:
        # Set bounds on x_i to be positive (choose a "large enough" value for the upper bound)
        bnds = ((0., 1000.),) * S.shape[1]

        # Find the x that minimizes this unconstrained ridge/lasso Regression problem:
        if formulation == 'ridge':
            soln = opt.minimize(compute_ridge, x0, bounds=bnds, constraints=None, args=(S,m,reg_constant))
        elif formulation == 'lasso':
            soln = opt.minimize(compute_lasso, x0, bounds=bnds, constraints=None, args=(S,m,reg_constant))

    if print_results:
        print()
        print(label)
        print(soln)
        print()
    else:
        if soln['success'] == False:
            warnings.warn("DeconRNASeq optimization did not terminate successfully. For more details: re-run with optional argument 'print_results:True'", category=UserWarning)

    xopt = soln['x']

    return xopt



def DeconRNASeq_main(rna_df, sig_df, patient_IDs='ALL', args={}):
    """
    This function does the following:
        - parses the dictionary 'args' for the arguments to pass on to the DeconRNASeq method.
        - eliminates genes from rna_df and sig_df that are not present in both data sets
        - Runs DeconRNASeq() for each patient specified in patient_IDs' argument
        - Combines the resulting frequencies into a pandas dataframe (num_celltypes x num_patients)
    Inputs:
        - rna_df: pandas df of rna gene expression data.
            Rows are genes (indexed by 'Hugo_Symbol') and columns are patients
        - sig_df: pandas df of Signature gene expression values for given cell types.
            Rows are genes (indexed by 'Hugo_Symbol') and columns are cell types
        - patient_IDs: list of patient IDs to run DeconRNASeq for.
                Alternatively, can use the string 'ALL' to run for all patients
        - args: dictionary containing any of the following:
            - check_sig: boolean, whether or not to check the condition number of the signature matrix
            - scaling: string, must be either 'None', 'zscore', or 'minmax'. Determines how to scale the signature matrix and mixture data before solving for optimal x
            - scaling_axis: 0 or 1. Whether to scale mixture data and signature matrix by normalizing each column (celltype/patient) separately (scaling_axis=0) or each row (gene) separately (scaling_axis=1).
            - formulation: see DeconRNASeq()
            - reg_constant: see DeconRNASeq()
            - print_result: see DeconRNASeq()
    Outputs:
        - cell_freqs: pandas df. Contains cell type frequencies for each patient in 'patient_IDs' list.
            Rows are indexed by cell type, columns are patient IDs
    """
    import pandas as pd
    import numpy as np
    from .data_utils import keep_common_genes
    from .data_utils import df_normalization

    # Read in optional arguments, or set them as default
    # Assert values are of the right data type when passed to DeconRNASeq() function.

    # formulation must be 'qp', 'ridge', or 'lasso'
    if 'formulation' in args.keys():
        formulation = args['formulation']
        if formulation not in ['qp','ridge','lasso']:
            raise ValueError("Formulation ({!r}) must be set to 'qp', 'ridge', or 'lasso'".format(formulation))
    else:
        formulation = 'qp'
    # reg_constant must be a double
    if 'reg_constant' in args.keys():
        reg_constant = args['reg_constant']
    else:
        reg_constant = 1.0
    if 'check_sig' in args.keys():
        check_sig = args['check_sig']
        if not isinstance(check_sig, bool):
            raise ValueError("check_sig ({!r}) must be a boolean variable".format(check_sig))
    else:
        check_sig = False
    if 'scaling' in args.keys():
        scaling = args['scaling']
        if scaling not in ['None', 'none', 'zscore', 'minmax', 'r-zscore']:
            raise ValueError("scaling ({!r}) must be set to 'none', 'zscore' or 'minmax'".format(scaling))
    else:
        scaling = 'minmax'
    if 'scaling_axis' in args.keys():
        scaling_axis = args['scaling_axis']
        if scaling_axis not in [0, 1]:
            raise ValueError("scaling_axis ({!r}) must be 0 or 1".format(scaling_axis))
    else:
        scaling_axis = 0
    if 'print_results' in args.keys():
        print_results = args['print_results']
        if not isinstance(print_results, bool):
            raise ValueError("print_results ({!r}) must be a boolean variable".format(print_results))
    else:
        print_results = False


    # eliminate genes not present in both rna and sig dfs, and ensure they are in the same order:
    rna_df, sig_df = keep_common_genes(rna_df, sig_df)

    # Scale Data:
    if scaling in ['zscore', 'minmax', 'r-zscore']:
        # R implementation uses zscore scaling.
        sig_df = df_normalization(sig_df, scaling=scaling, axis=scaling_axis)
        rna_df = df_normalization(rna_df, scaling=scaling, axis=scaling_axis)

    # Convert signature to numpy array
    Sig = np.array(sig_df)

    # Check the condition number of the signature matrix:
    if check_sig:
        print("Condition number of signature matrix =", np.linalg.cond(Sig))

    # Select a patient / list of patients to solve for their cell type frequencies:
         # Patient_ID must be 'ALL' or an array of specific patient IDs.
    if patient_IDs == 'ALL':
        patient_list = rna_df.columns
    elif not isinstance(patient_IDs, type([])):
        raise ValueError("patient_IDs should be either 'ALL', or an array of IDs (not a single ID)")
    else:
        patient_list = patient_IDs

    # For each patient, run DeconRNASeq to get cell type frequencies, and save results to pandas df:
    print("Running DeconRNASeq...")
    cell_freqs_df = pd.DataFrame()
    cell_freqs_df['Patient_ID'] = sig_df.columns
    cell_freqs_df = cell_freqs_df.set_index(['Patient_ID'])
    for patient in patient_list:
        if patient in rna_df.columns:
            Mix = np.array(rna_df[patient])
            cell_freqs_df[patient] = DeconRNASeq(Sig, Mix, formulation=formulation, reg_constant=reg_constant, print_results=print_results, label=patient)
        else:
            raise ValueError("patient_ID ({!r}) not present in rna dataframe".format(patient))

    cell_freqs_df = cell_freqs_df.transpose()

    return cell_freqs_df
